use rand::rngs::StdRng;
use rand::RngCore;
use rand::SeedableRng;
use rug::{Assign, Integer};

pub fn encrypt_bit(bit: u8, key: &Integer) -> Integer {
    assert!(bit == 0 || bit == 1, "Only bits (0 or 1) are allowed");

    let mut rng = StdRng::from_os_rng();

    let q: Integer = Integer::from(rng.next_u64()) % key;

    // noise
    let p_div_4: Integer = key.clone() / 4;
    let r = std::iter::repeat_with(|| Integer::from(rng.next_u64()))
        .find(|r| r.clone() * 2 < p_div_4)
        .unwrap();

    // c = b + p*q + 2*r
    let mut c = Integer::new();
    c.assign(key.clone() * &q + 2 * &r + bit);

    c
}

pub fn decrypt_bit(encrypted: &Integer, key: &Integer) -> u8 {
    let bit: Integer = (encrypted.clone() % key.clone()) % 2; // ((c mod p) mod 2)
    bit.to_u8().unwrap()
}

pub fn generate_secret_key(eta: u32) -> Integer {
    let mut rng = StdRng::from_os_rng();
    let lower = Integer::from(1) << (eta - 1);
    let upper = Integer::from(1) << eta;
    let range: Integer = upper - &lower;
    std::iter::repeat_with(|| Integer::from(rng.next_u64()) % &range + &lower)
        .find(|p: &Integer| p.is_odd())
        .unwrap()
}
